#!/usr/bin/python
"""
This file is needed to decode the command line string sent from the BPS
plugin -> PanDA -> Edge node cluster management
-> Edge node -> Container. This file is not a part
of the BPS but a part of the payload wrapper.
It decodes the hexified command line.
"""
import binascii
import os
import re
import sys

from lsst.resources import ResourcePath


def replace_placeholders(cmd_line, tag, replancements):
    occurences_to_replace = re.findall(f"<{tag}:(.*?)>", cmd_line)
    for placeholder in occurences_to_replace:
        if placeholder in replancements:
            cmd_line = cmd_line.replace(f"<{tag}:{placeholder}>", replancements[placeholder])
        else:
            raise ValueError(
                "ValueError exception thrown, because "
                f"{placeholder} is not found in the "
                "replacement values and could "
                "not be passed to the command line"
            )
    return cmd_line


def replace_environment_vars(cmd_line):
    """Replaces placeholders to the actual environment variables.

    Parameters
    ----------
    cmd_line : `str`
        Command line

    Returns
    -------
    cmdline: `str`
        Processed command line
    """
    environment_vars = os.environ
    cmd_line = replace_placeholders(cmd_line, "ENV", environment_vars)
    return cmd_line


def replace_files_placeholders(cmd_line, files):
    """Replaces placeholders for files.

    Parameters
    ----------
    cmd_line : `str`
        Command line
    files : `str`
        String with key:value pairs separated by the '+' sign.
        Keys contain the file type (runQgraphFile,
        butlerConfig, ...).
        Values contains file names.

    Returns
    -------
    cmd_line: `str`
        Processed command line
    """

    files_key_vals = files.split("+")
    files = {}
    for file in files_key_vals:
        file_name_placeholder, file_name = file.split(":")
        files[file_name_placeholder] = file_name
    cmd_line = replace_placeholders(cmd_line, "FILE", files)
    return cmd_line


def deliver_input_files(src_path, files, skip_copy):
    """Delivers input files needed for a job

    Parameters
    ----------
    src_path : `str`
        URI for folder where the input files placed
    files : `str`
        String with file names separated by the '+' sign

    Returns
    -------
    cmdline: `str`
        Processed command line
        :param skip_copy:
    """

    files = files.split("+")
    src_uri = ResourcePath(src_path, forceDirectory=True)
    for file in files:
        file_name_placeholder, file_pfn = file.split(":")
        if file_name_placeholder not in skip_copy.split("+"):
            src = src_uri.join(file_pfn)
            base_dir = None
            if src.isdir():
                files_to_copy = ResourcePath.findFileResources([src])
                base_dir = file_pfn
            else:
                files_to_copy = [src]
            dest_base = ResourcePath("", forceAbsolute=True, forceDirectory=True)
            if base_dir:
                dest_base = dest_base.join(base_dir)
            for file_to_copy in files_to_copy:
                dest = dest_base.join(file_to_copy.basename())
                dest.transfer_from(file_to_copy, transfer="copy")
                print(f"copied {file_to_copy.path} to {dest.path}", file=sys.stderr)
            if file_name_placeholder == "job_executable":
                os.chmod(dest.path, 0o777)


deliver_input_files(sys.argv[3], sys.argv[4], sys.argv[5])
cmd_line = str(binascii.unhexlify(sys.argv[1]).decode())
data_params = sys.argv[2].split("+")
cmd_line = replace_environment_vars(cmd_line)

"""This call replaces the pipetask command line placeholders
 with actual data provided in the script call
 in form placeholder1:file1+placeholder2:file2:...
"""
cmd_line = replace_files_placeholders(cmd_line, sys.argv[4])

for key_value_pair in data_params[1:]:
    (key, value) = key_value_pair.split(":")
    cmd_line = cmd_line.replace("{" + key + "}", value)

print(cmd_line)

exit_status = os.system(cmd_line)
exit_code = 1
if os.WIFSIGNALED(exit_status):
    exit_code = os.WTERMSIG(exit_status) + 128
elif os.WIFEXITED(exit_status):
    exit_code = os.WEXITSTATUS(exit_status)
sys.exit(exit_code)



# This file is part of ctrl_bps.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Subcommand definitions for the PanDA auth commands.
"""

__all__ = [
    "status",
    "reset",
    "clean",
]


import click
from lsst.daf.butler.cli.utils import MWCommand

from ...panda_auth_drivers import panda_auth_clean_driver, panda_auth_reset_driver, panda_auth_status_driver


class PandaAuthCommand(MWCommand):
    """Command subclass with panda-auth-command specific overrides."""

    extra_epilog = "See 'panda_auth --help' for more options."


@click.command(cls=PandaAuthCommand)
def status(*args, **kwargs):
    """Print informatino about auth token."""
    panda_auth_status_driver(*args, **kwargs)


@click.command(cls=PandaAuthCommand)
def reset(*args, **kwargs):
    """Get new auth token."""
    panda_auth_reset_driver(*args, **kwargs)


@click.command(cls=PandaAuthCommand)
def clean(*args, **kwargs):
    """Clean up token and token cache files."""
    panda_auth_clean_driver(*args, **kwargs)



# This file is part of ctrl_bps_panda.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Functions for each panda_auth subcommand.
"""


__all__ = [
    "panda_auth_clean",
    "panda_auth_expiration",
    "panda_auth_setup",
    "panda_auth_status",
    "panda_auth_update",
]


import logging
import os

import idds.common.utils as idds_utils
import pandaclient.idds_api
from pandaclient.openidc_utils import OpenIdConnect_Utils

_LOG = logging.getLogger(__name__)


def panda_auth_clean():
    """Clean up token and token cache files."""
    open_id = panda_auth_setup()
    open_id.cleanup()


def panda_auth_expiration():
    """Get number of seconds until token expires.

    Return
    ------
    expiration : `int`
        Number of seconds until token expires.
    """
    expiration = 0
    ret = panda_auth_status()
    if ret:
        expiration = ret[-1]["exp"]
    return expiration


def panda_auth_setup():
    """Initialize auth object used by various auth functions.

    Return
    ------
    open_id : `pandaclient.openidc_utils.OpenIdConnect_Utils`
        Auth object which can interact with auth token.
    """
    for key in [
        "PANDA_AUTH",
        "PANDA_VERIFY_HOST",
        "PANDA_AUTH_VO",
        "PANDA_URL_SSL",
        "PANDA_URL",
    ]:
        if key not in os.environ:
            raise OSError(f"Missing environment variable {key}")

    # OpenIdConnect_Utils have a verbose flag that filters
    # some debugging messages.  If user chose debug, just
    # turn on all of the messages.
    verbose = False
    if _LOG.isEnabledFor(logging.DEBUG):
        verbose = True

    open_id = OpenIdConnect_Utils(None, log_stream=_LOG, verbose=verbose)
    return open_id


def panda_auth_status():
    """Gather information about a token if it exists.

    Return
    ------
    status : `dict`
        Status information about a token if it exists.
        Includes filename and expiration epoch.
    """
    status = None
    open_id = panda_auth_setup()
    ret = open_id.check_token()
    if ret and ret[0]:
        # get_token_path will return the path even if a token doesn't
        # currently exist.  So check for token first via check_token, then
        # add path.
        status = {"filename": open_id.get_token_path()}
        status.update(ret[-1])
    return status


def panda_auth_update(idds_server=None, reset=False):
    """Get new auth token if needed or reset is True.

    Parameters
    ----------
    idds_server : `str`, optional
        URL for the iDDS server.  Defaults to None which means that the
        underlying functions use any value in the panda relay service.
    reset : `bool`, optional
        Whether to first clean up any previous token.  Defaults to False.
    """
    if reset:
        panda_auth_clean()

    # Create client manager
    # (There is a function in OpenIdConnect_Utils, but it takes several
    #  parameters.  Letting the client manager do it is currently easiest
    #  way to match what happens when the workflow is actually submitted.)
    cm = pandaclient.idds_api.get_api(
        idds_utils.json_dumps, idds_host=idds_server, compress=True, manager=True, verbose=False
    )

    # Must call some function to actually check auth
    # https://panda-wms.readthedocs.io/en/latest/client/notebooks/jupyter_setup.html#Get-an-OIDC-ID-token
    ret = cm.get_status(request_id=0, with_detail=False)
    _LOG.debug("get_status results: %s", ret)

    # Check success
    # https://panda-wms.readthedocs.io/en/latest/client/rest_idds.html
    if ret[0] == 0 and ret[1][0]:
        # The success keys from get_status currently do not catch if invalid
        # idds server given.  So for now, check result string for keywords.
        if "request_id" not in ret[1][-1] or "status" not in ret[1][-1]:
            raise RuntimeError(f"Error contacting PanDA service: {str(ret)}")



# This file is part of ctrl_bps_panda.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
import os
import re

_LOG = logging.getLogger(__name__)


class CommandLineEmbedder:
    """
    Class embeds static (constant across a task) values
    into the pipeline execution command line
    and resolves submission side environment variables

    Parameters
    ----------
    config : `lsst.ctrl.bps.BpsConfig`
            BPS configuration that includes the list of dynamic
            (uniques per job) and submission side resolved variables

    """

    def __init__(self, config):
        self.leave_placeholder_params = config.get("placeholderParams", ["qgraphNodeId", "qgraphId"])
        self.submit_side_resolved = config.get("submitSideResolvedParams", ["USER"])

    def replace_static_parameters(self, cmd_line, lazy_vars):
        """Substitutes the lazy parameters in the command line which
        are static, the same for every job in the workflow and could be
        defined once. This function offloads the edge node processing
        and number of parameters transferred together with job


        Parameters
        ----------
        cmd_line: `str` command line to be processed
        lazy_vars: `dict` of lazy variables and its values

        Returns
        -------
        processed command line
        """
        for param_name, param_val in lazy_vars.items():
            if param_name not in self.leave_placeholder_params:
                cmd_line = cmd_line.replace("{" + param_name + "}", param_val)
        return cmd_line

    def resolve_submission_side_env_vars(self, cmd_line):
        """Substitutes the lazy parameters in the command line
        which are defined and resolved on the submission side

        Parameters
        ----------
        cmd_line: `str` command line to be processed

        Returns
        -------
        processed command line
        """

        for param in self.submit_side_resolved:
            if os.getenv(param):
                cmd_line = cmd_line.replace("<ENV:" + param + ">", os.getenv(param))
            else:
                _LOG.info("Expected parameter {0} is not found in the environment variables".format(param))
        return cmd_line

    def attach_pseudo_file_params(self, lazy_vars):
        """Adds the parameters needed to finalize creation of a pseudo file

        Parameters
        ----------
        lazy_vars: `dict` of values of to be substituted

        Returns
        -------
        pseudo input file name suffix
        """

        file_suffix = ""
        for item in self.leave_placeholder_params:
            file_suffix += "+" + item + ":" + lazy_vars.get(item, "")
        return file_suffix

    def substitute_command_line(self, cmd_line, lazy_vars, job_name):
        """Preprocesses the command line leaving for the egde node evaluation
        only parameters which are job / environment dependent

        Parameters
        ----------
        bps_file_name: `str` input file name proposed by BPS
        cmd_line: `str` command line containing all lazy placeholders
        lazy_vars: `dict` of lazy parameter name/values
        job_name: `str` job name proposed by BPS

        Returns
        -------
        cmd_line: `str`
            processed command line
        file_name: `str`
            job pseudo input file name
        """
        cmd_vals = set([m.group(1) for m in re.finditer(r"[^$]{([^}]+)}", cmd_line)])
        actual_lazy_vars = {}
        for key in cmd_vals:
            actual_lazy_vars[key] = lazy_vars[key]

        cmd_line = self.replace_static_parameters(cmd_line, actual_lazy_vars)
        cmd_line = self.resolve_submission_side_env_vars(cmd_line)
        file_name = job_name + self.attach_pseudo_file_params(actual_lazy_vars)
        return cmd_line, file_name



# This file is part of ctrl_bps_panda.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Driver functions for each panda_auth subcommand.

Driver functions ensure that ensure all setup work is done before running
the subcommand method.
"""


__all__ = [
    "panda_auth_clean_driver",
    "panda_auth_reset_driver",
    "panda_auth_status_driver",
]


import logging
from datetime import datetime

from .panda_auth_utils import panda_auth_clean, panda_auth_status, panda_auth_update

_LOG = logging.getLogger(__name__)


def panda_auth_clean_driver():
    """Clean up token and token cache files."""
    panda_auth_clean()


def panda_auth_reset_driver():
    """Get new auth token."""
    panda_auth_update(None, True)


def panda_auth_status_driver():
    """Gather information about a token if it exists."""
    status = panda_auth_status()
    if status:
        print(f"{'Filename:':15} {status['filename']}")
        print(f"{'Valid starting:':15} {datetime.fromtimestamp(status['iat'])}")
        print(f"{'Expires:':15} {datetime.fromtimestamp(status['exp'])}")
        print(f"{'Name:':15} {status['name']}")
        print(f"{'Email:':15} {status['email']}")
        print(f"{'Groups:':15} {','.join(status['groups'])}")
        print(f"{'Organization:':15} {status['organisation_name']}")
    else:
        print("No token found")



# This file is part of ctrl_bps_panda.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Symbolic constants
"""

PANDA_DEFAULT_PROD_SOURCE_LABEL = "managed"
PANDA_DEFAULT_PROCESSING_TYPE = None
PANDA_DEFAULT_TASK_TYPE = "test"
PANDA_DEFAULT_VO = "wlcg"
PANDA_DEFAULT_PRIORITY = 900
PANDA_DEFAULT_MAX_JOBS_PER_TASK = 30000
PANDA_DEFAULT_RSS = 1024
PANDA_DEFAULT_MAX_WALLTIME = 90000
PANDA_DEFAULT_MAX_ATTEMPTS = 3
PANDA_DEFAULT_CORE_COUNT = 1
PANDA_DEFAULT_CLOUD = "US"
PANDA_DEFAULT_MAX_COPY_WORKERS = 10
PANDA_MAX_LEN_INPUT_FILE = 4000



"""Unit tests for PanDA authentication utilities.
"""
import os
import unittest
from unittest import mock

from lsst.ctrl.bps.panda import __version__ as version
from lsst.ctrl.bps.panda.panda_auth_utils import panda_auth_status


class VersionTestCase(unittest.TestCase):
    def test_version(self):
        # Check that version is defined.
        self.assertIsNotNone(version)


class TestPandaAuthUtils(unittest.TestCase):
    def testPandaAuthStatusWrongEnviron(self):
        unwanted = {
            "PANDA_AUTH",
            "PANDA_VERIFY_HOST",
            "PANDA_AUTH_VO",
            "PANDA_URL_SSL",
            "PANDA_URL",
        }
        test_environ = {key: val for key, val in os.environ.items() if key not in unwanted}
        with mock.patch.dict(os.environ, test_environ, clear=True):
            with self.assertRaises(OSError):
                panda_auth_status()


if __name__ == "__main__":
    unittest.main()



# This file is part of ctrl_bps_panda.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import click
from lsst.daf.butler.cli.butler import LoaderCLI
from lsst.daf.butler.cli.opt import (
    log_file_option,
    log_label_option,
    log_level_option,
    log_tty_option,
    long_log_option,
)


class PandaAuthCli(LoaderCLI):
    localCmdPkg = "lsst.ctrl.bps.panda.cli.cmd"


@click.command(cls=PandaAuthCli, context_settings=dict(help_option_names=["-h", "--help"]), epilog=None)
@log_level_option(default=["INFO"])
@long_log_option()
@log_file_option()
@log_tty_option()
@log_label_option()
def cli(log_level, long_log, log_file, log_tty, log_label):
    pass


def main():
    return cli()



# This file is part of ctrl_bps_panda.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Interface between generic workflow to PanDA/iDDS workflow system.
"""


__all__ = ["PanDAService", "PandaBpsWmsWorkflow"]


import json
import logging
import os
import pickle
import re

from idds.workflowv2.workflow import Workflow as IDDS_client_workflow
from lsst.ctrl.bps.panda.constants import PANDA_DEFAULT_MAX_COPY_WORKERS
from lsst.ctrl.bps.panda.utils import (
    add_final_idds_work,
    add_idds_work,
    copy_files_for_distribution,
    get_idds_client,
    get_idds_result,
)
from lsst.ctrl.bps.wms_service import BaseWmsService, BaseWmsWorkflow, WmsRunReport, WmsStates

_LOG = logging.getLogger(__name__)


class PanDAService(BaseWmsService):
    """PanDA version of WMS service"""

    def prepare(self, config, generic_workflow, out_prefix=None):
        # Docstring inherited from BaseWmsService.prepare.
        _LOG.debug("out_prefix = '%s'", out_prefix)
        workflow = PandaBpsWmsWorkflow.from_generic_workflow(
            config, generic_workflow, out_prefix, f"{self.__class__.__module__}.{self.__class__.__name__}"
        )
        workflow.write(out_prefix)
        return workflow

    def submit(self, workflow):
        _, max_copy_workers = self.config.search(
            "maxCopyWorkers", opt={"default": PANDA_DEFAULT_MAX_COPY_WORKERS}
        )
        # Docstring inherited from BaseWmsService.submit.
        copy_files_for_distribution(
            workflow.files_to_pre_stage, self.config["fileDistributionEndPoint"], max_copy_workers
        )

        idds_client = get_idds_client(self.config)
        ret = idds_client.submit(workflow.idds_client_workflow, username=None, use_dataset_name=False)
        _LOG.debug("iDDS client manager submit returned = %s", ret)

        # Check submission success
        status, result, error = get_idds_result(ret)
        if status:
            request_id = int(result)
        else:
            raise RuntimeError(f"Error submitting to PanDA service: {error}")

        _LOG.info("Submitted into iDDs with request id=%s", request_id)
        workflow.run_id = request_id

    def restart(self, wms_workflow_id):
        # Docstring inherited from BaseWmsService.restart.
        idds_client = get_idds_client(self.config)
        ret = idds_client.retry(request_id=wms_workflow_id)
        _LOG.debug("Restart PanDA workflow returned = %s", ret)

        status, result, error = get_idds_result(ret)
        if status:
            _LOG.info("Restarting PanDA workflow %s", result)
            return wms_workflow_id, None, json.dumps(result)

        return None, None, f"Error retry PanDA workflow: {str(error)}"

    def report(self, wms_workflow_id=None, user=None, hist=0, pass_thru=None, is_global=False):
        # Docstring inherited from BaseWmsService.report.
        message = ""
        run_reports = []

        if not wms_workflow_id:
            message = "Run summary not implemented yet, use 'bps report --id <workflow_id>' instead"
            return run_reports, message

        idds_client = get_idds_client(self.config)
        ret = idds_client.get_requests(request_id=wms_workflow_id, with_detail=True)
        _LOG.debug("PanDA get workflow status returned = %s", str(ret))

        request_status = ret[0]
        if request_status != 0:
            raise RuntimeError(f"Error to get workflow status: {ret} for id: {wms_workflow_id}")

        tasks = ret[1][1]
        if not tasks:
            message = f"No records found for workflow id '{wms_workflow_id}'. Hint: double check the id"
        else:
            head = tasks[0]
            wms_report = WmsRunReport(
                wms_id=str(head["request_id"]),
                operator=head["username"],
                project="",
                campaign="",
                payload="",
                run=head["name"],
                state=WmsStates.UNKNOWN,
                total_number_jobs=0,
                job_state_counts={state: 0 for state in WmsStates},
                job_summary={},
                run_summary="",
            )

            # The status of a task is taken from the first item of state_map.
            # The workflow is in status WmsStates.FAILED when:
            #      All tasks have failed.
            # SubFinished tasks has jobs in
            #      output_processed_files: Finished
            #      output_failed_files: Failed
            #      output_missing_files: Missing
            state_map = {
                "Finished": [WmsStates.SUCCEEDED],
                "SubFinished": [
                    WmsStates.SUCCEEDED,
                    WmsStates.FAILED,
                    WmsStates.PRUNED,
                ],
                "Transforming": [
                    WmsStates.RUNNING,
                    WmsStates.SUCCEEDED,
                    WmsStates.FAILED,
                    WmsStates.UNREADY,
                    WmsStates.PRUNED,
                ],
                "Failed": [WmsStates.FAILED, WmsStates.PRUNED],
            }

            file_map = {
                WmsStates.SUCCEEDED: "output_processed_files",
                WmsStates.RUNNING: "output_processing_files",
                WmsStates.FAILED: "output_failed_files",
                WmsStates.UNREADY: "input_new_files",
                WmsStates.PRUNED: "output_missing_files",
            }

            # workflow status to report as SUCCEEDED
            wf_status = ["Finished", "SubFinished", "Transforming"]

            wf_succeed = False

            tasks.sort(key=lambda x: x["transform_workload_id"])

            # Loop over all tasks data returned by idds_client
            for task in tasks:
                totaljobs = task["output_total_files"]
                wms_report.total_number_jobs += totaljobs
                tasklabel = task["transform_name"]
                tasklabel = re.sub(wms_report.run + "_", "", tasklabel)
                status = task["transform_status"]["attributes"]["_name_"]
                taskstatus = {}
                # Fill number of jobs in all WmsStates
                for state in WmsStates:
                    njobs = 0
                    # Each WmsState have many iDDS status mapped to it.
                    for mappedstate in state_map[status]:
                        if state in file_map and mappedstate == state:
                            if task[file_map[mappedstate]] is not None:
                                njobs = task[file_map[mappedstate]]
                            if state == WmsStates.RUNNING:
                                njobs += task["output_new_files"] - task["input_new_files"]
                            break
                    wms_report.job_state_counts[state] += njobs
                    taskstatus[state] = njobs
                wms_report.job_summary[tasklabel] = taskstatus

                # To fill the EXPECTED column
                if wms_report.run_summary:
                    wms_report.run_summary += ";"
                wms_report.run_summary += f"{tasklabel}:{str(totaljobs)}"

                if status in wf_status:
                    wf_succeed = True
                    wms_report.state = state_map[status][0]

            # All tasks have failed, set the workflow FAILED
            if not wf_succeed:
                wms_report.state = WmsStates.FAILED

            run_reports.append(wms_report)

        return run_reports, message

    def list_submitted_jobs(self, wms_id=None, user=None, require_bps=True, pass_thru=None, is_global=False):
        # Docstring inherited from BaseWmsService.list_submitted_jobs.
        if wms_id is None and user is not None:
            raise RuntimeError(
                "Error to get workflow status report: wms_id is required"
                " and filtering workflows with 'user' is not supported."
            )

        idds_client = get_idds_client(self.config)
        ret = idds_client.get_requests(request_id=wms_id)
        _LOG.debug("PanDA get workflows returned = %s", ret)

        status, result, error = get_idds_result(ret)
        if status:
            req_ids = [req["request_id"] for req in result]
            return req_ids

        raise RuntimeError(f"Error list PanDA workflow requests: {error}")

    def cancel(self, wms_id, pass_thru=None):
        # Docstring inherited from BaseWmsService.cancel.
        idds_client = get_idds_client(self.config)
        ret = idds_client.abort(request_id=wms_id)
        _LOG.debug("Abort PanDA workflow returned = %s", ret)

        status, result, error = get_idds_result(ret)
        if status:
            _LOG.info("Aborting PanDA workflow %s", result)
            return True, json.dumps(result)

        return False, f"Error abort PanDA workflow: {str(error)}"

    def ping(self, pass_thru=None):
        # Docstring inherited from BaseWmsService.ping.
        idds_client = get_idds_client(self.config)
        ret = idds_client.ping()
        _LOG.debug("Ping PanDA service returned = %s", ret)

        status, result, error = get_idds_result(ret)
        if status:
            if "Status" in result and result["Status"] == "OK":
                return 0, None

            return -1, f"Error ping PanDA service: {str(result)}"

        return -1, f"Error ping PanDA service: {str(error)}"

    def run_submission_checks(self):
        # Docstring inherited from BaseWmsService.run_submission_checks.
        for key in ["PANDA_URL"]:
            if key not in os.environ:
                raise OSError(f"Missing environment variable {key}")

        status, message = self.ping()
        if status != 0:
            raise RuntimeError(message)


class PandaBpsWmsWorkflow(BaseWmsWorkflow):
    """A single Panda based workflow

    Parameters
    ----------
    name : `str`
        Unique name for Workflow
    config : `lsst.ctrl.bps.BpsConfig`
        BPS configuration that includes necessary submit/runtime information
    """

    def __init__(self, name, config=None):
        super().__init__(name, config)
        self.files_to_pre_stage = {}  # src, dest
        self.idds_client_workflow = IDDS_client_workflow(name=name)

    @classmethod
    def from_generic_workflow(cls, config, generic_workflow, out_prefix, service_class):
        # Docstring inherited from BaseWmsWorkflow.from_generic_workflow.
        wms_workflow = cls(generic_workflow.name, config)

        files, dag_sink_work, task_count = add_idds_work(
            config, generic_workflow, wms_workflow.idds_client_workflow
        )
        wms_workflow.files_to_pre_stage.update(files)

        files = add_final_idds_work(
            config, generic_workflow, wms_workflow.idds_client_workflow, dag_sink_work, task_count + 1, 1
        )
        wms_workflow.files_to_pre_stage.update(files)

        return wms_workflow

    def write(self, out_prefix):
        # Docstring inherited from BaseWmsWorkflow.write.
        with open(os.path.join(out_prefix, "panda_workflow.pickle"), "wb") as fh:
            pickle.dump(self, fh)



# This file is part of ctrl_bps_panda.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Utilities for bps PanDA plugin."""

__all__ = [
    "copy_files_for_distribution",
    "get_idds_client",
    "get_idds_result",
    "convert_exec_string_to_hex",
    "add_decoder_prefix",
]

import binascii
import concurrent.futures
import logging
import os

import idds.common.utils as idds_utils
import pandaclient.idds_api
from idds.doma.workflowv2.domapandawork import DomaPanDAWork
from idds.workflowv2.workflow import AndCondition
from lsst.ctrl.bps import BpsConfig, GenericWorkflow, GenericWorkflowJob
from lsst.ctrl.bps.panda.cmd_line_embedder import CommandLineEmbedder
from lsst.ctrl.bps.panda.constants import (
    PANDA_DEFAULT_CLOUD,
    PANDA_DEFAULT_CORE_COUNT,
    PANDA_DEFAULT_MAX_ATTEMPTS,
    PANDA_DEFAULT_MAX_JOBS_PER_TASK,
    PANDA_DEFAULT_MAX_WALLTIME,
    PANDA_DEFAULT_PRIORITY,
    PANDA_DEFAULT_PROCESSING_TYPE,
    PANDA_DEFAULT_PROD_SOURCE_LABEL,
    PANDA_DEFAULT_RSS,
    PANDA_DEFAULT_TASK_TYPE,
    PANDA_DEFAULT_VO,
)
from lsst.resources import ResourcePath

_LOG = logging.getLogger(__name__)


def copy_files_for_distribution(files_to_stage, file_distribution_uri, max_copy_workers):
    """Brings locally generated files into Cloud for further
    utilization them on the edge nodes.

    Parameters
    ----------
    local_pfns : `dict` [`str`, `str`]
        Files which need to be copied to a workflow staging area.
    file_distribution_uri: `str`
        Path on the edge node accessed storage,
        including access protocol, bucket name to place files.
    max_copy_workers : `int`
        Maximum number of workers for copying files.

    Raises
    ------
    RuntimeError
        Raised when error copying files to the distribution point.
    """
    files_to_copy = {}

    # In case there are folders we iterate over its content
    for local_pfn in files_to_stage.values():
        folder_name = os.path.basename(os.path.normpath(local_pfn))
        if os.path.isdir(local_pfn):
            files_in_folder = ResourcePath.findFileResources([local_pfn])
            for file in files_in_folder:
                file_name = file.basename()
                files_to_copy[file] = ResourcePath(
                    os.path.join(file_distribution_uri, folder_name, file_name)
                )
        else:
            files_to_copy[ResourcePath(local_pfn)] = ResourcePath(
                os.path.join(file_distribution_uri, folder_name)
            )

    copy_executor = concurrent.futures.ThreadPoolExecutor(max_workers=max_copy_workers)
    future_file_copy = []
    for src, trgt in files_to_copy.items():
        # S3 clients explicitly instantiate here to overpass this
        # https://stackoverflow.com/questions/52820971/is-boto3-client-thread-safe
        trgt.exists()
        future_file_copy.append(copy_executor.submit(trgt.transfer_from, src, transfer="copy"))

    for future in concurrent.futures.as_completed(future_file_copy):
        if not future.result() is None:
            raise RuntimeError("Error of placing files to the distribution point")


def get_idds_client(config):
    """Get the idds client.

    Parameters
    ----------
    config : `lsst.ctrl.bps.BpsConfig`
        BPS configuration.

    Returns
    -------
    idds_client: `idds.client.clientmanager.ClientManager`
        iDDS ClientManager object.
    """
    idds_server = None
    if isinstance(config, BpsConfig):
        _, idds_server = config.search("iddsServer", opt={"default": None})
    elif isinstance(config, dict) and "iddsServer" in config:
        idds_server = config["iddsServer"]
    # if idds_server is None, a default value on the panda relay service
    # will be used
    idds_client = pandaclient.idds_api.get_api(
        idds_utils.json_dumps, idds_host=idds_server, compress=True, manager=True
    )
    return idds_client


def get_idds_result(ret):
    """Parse the results returned from iDDS.

    Parameters
    ----------
    ret: `tuple` of (`int`, (`bool`, payload)).
        The first part ret[0] is the status of PanDA relay service.
        The part of ret[1][0] is the status of iDDS service.
        The part of ret[1][1] is the returned payload.
        If ret[1][0] is False, ret[1][1] can be error messages.

    Returns
    -------
    status: `bool`
        The status of iDDS calls.
    result: `int` or `list` or `dict`
        The result returned from iDDS.
    error: `str`
        Error messages.
    """
    # https://panda-wms.readthedocs.io/en/latest/client/rest_idds.html
    if not isinstance(ret, (list, tuple)) or ret[0] != 0:
        # Something wrong with the PanDA relay service.
        # The call may not be delivered to iDDS.
        status = False
        result = None
        error = f"PanDA relay service returns errors: {str(ret)}"
    else:
        if ret[1][0]:
            status = True
            result = ret[1][1]
            error = None
            if isinstance(result, str) and "Authentication no permission" in result:
                status = False
                result = None
                error = result
        else:
            # iDDS returns errors
            status = False
            result = None
            error = f"iDDS returns errors: {str(ret[1][1])}"
    return status, result, error


def _make_pseudo_filename(config, gwjob):
    """Make the job pseudo filename.

    Parameters
    ----------
    config : `lsst.ctrl.bps.BpsConfig`
        BPS configuration.
    gwjob : `lsst.ctrl.bps.GenericWorkflowJob`
        Job for which to create the pseudo filename.

    Returns
    -------
    pseudo_filename : `str`
        The pseudo filename for the given job.
    """
    cmd_line_embedder = CommandLineEmbedder(config)
    _, pseudo_filename = cmd_line_embedder.substitute_command_line(
        gwjob.executable.src_uri + " " + gwjob.arguments, gwjob.cmdvals, gwjob.name
    )
    return pseudo_filename


def _make_doma_work(config, generic_workflow, gwjob, task_count, task_chunk):
    """Make the DOMA Work object for a PanDA task.

    Parameters
    ----------
    config : `lsst.ctrl.bps.BpsConfig`
        BPS configuration.
    gwjob : `lsst.ctrl.bps.GenericWorkflowJob`
        Job representing the jobs for the PanDA task.
    task_count : `int`
        Count of PanDA tasks used when making unique names.
    task_chunk : `int`
        Count of chunk of a PanDA tasks used when making unique names.

    Returns
    -------
    work : `idds.doma.workflowv2.domapandawork.DomaPanDAWork`
        The client representation of a PanDA task.
    local_pfns : `dict` [`str`, `str`]
        Files which need to be copied to a workflow staging area.
    """
    _LOG.debug("Using gwjob %s to create new PanDA task (gwjob=%s)", gwjob.name, gwjob)
    cvals = {"curr_cluster": gwjob.label}
    _, site = config.search("computeSite", opt={"curvals": cvals, "required": True})
    cvals["curr_site"] = site
    _, processing_type = config.search(
        "processing_type", opt={"curvals": cvals, "default": PANDA_DEFAULT_PROCESSING_TYPE}
    )
    _, task_type = config.search("taskType", opt={"curvals": cvals, "default": PANDA_DEFAULT_TASK_TYPE})
    _, prod_source_label = config.search(
        "prodSourceLabel", opt={"curvals": cvals, "default": PANDA_DEFAULT_PROD_SOURCE_LABEL}
    )
    _, vo = config.search("vo", opt={"curvals": cvals, "default": PANDA_DEFAULT_VO})

    _, file_distribution_end_point = config.search(
        "fileDistributionEndPoint", opt={"curvals": cvals, "default": None}
    )

    # Assume input files are same across task
    local_pfns = {}
    direct_io_files = set()

    if gwjob.executable.transfer_executable:
        local_pfns["job_executable"] = gwjob.executable.src_uri
        job_executable = f"./{os.path.basename(gwjob.executable.src_uri)}"
    else:
        job_executable = gwjob.executable.src_uri
    cmd_line_embedder = CommandLineEmbedder(config)
    cmd_line, _ = cmd_line_embedder.substitute_command_line(
        job_executable + " " + gwjob.arguments, gwjob.cmdvals, gwjob.name
    )

    for gwfile in generic_workflow.get_job_inputs(gwjob.name, transfer_only=True):
        local_pfns[gwfile.name] = gwfile.src_uri
        if os.path.isdir(gwfile.src_uri):
            # this is needed to make isdir function working
            # properly in ButlerURL instance on the edge node
            local_pfns[gwfile.name] += "/"

        if gwfile.job_access_remote:
            direct_io_files.add(gwfile.name)

    if not direct_io_files:
        direct_io_files.add("cmdlineplaceholder")

    executable = add_decoder_prefix(
        config, cmd_line, file_distribution_end_point, (local_pfns, direct_io_files)
    )
    work = DomaPanDAWork(
        executable=executable,
        primary_input_collection={
            "scope": "pseudo_dataset",
            "name": f"pseudo_input_collection#{str(task_count)}",
        },
        output_collections=[
            {"scope": "pseudo_dataset", "name": f"pseudo_output_collection#{str(task_count)}"}
        ],
        log_collections=[],
        dependency_map=[],
        task_name=f"{generic_workflow.name}_{task_count:02d}_{gwjob.label}_{task_chunk:02d}",
        task_queue=gwjob.queue,
        task_log={
            "destination": "local",
            "value": "log.tgz",
            "dataset": "PandaJob_#{pandaid}/",
            "token": "local",
            "param_type": "log",
            "type": "template",
        },
        encode_command_line=True,
        task_rss=gwjob.request_memory if gwjob.request_memory else PANDA_DEFAULT_RSS,
        task_cloud=gwjob.compute_cloud if gwjob.compute_cloud else PANDA_DEFAULT_CLOUD,
        task_site=site,
        task_priority=int(gwjob.priority) if gwjob.priority else PANDA_DEFAULT_PRIORITY,
        core_count=gwjob.request_cpus if gwjob.request_cpus else PANDA_DEFAULT_CORE_COUNT,
        working_group=gwjob.accounting_group,
        processing_type=processing_type,
        task_type=task_type,
        prodSourceLabel=prod_source_label,
        vo=vo,
        maxattempt=gwjob.number_of_retries if gwjob.number_of_retries else PANDA_DEFAULT_MAX_ATTEMPTS,
        maxwalltime=gwjob.request_walltime if gwjob.request_walltime else PANDA_DEFAULT_MAX_WALLTIME,
    )
    return work, local_pfns


def add_final_idds_work(
    config, generic_workflow, idds_client_workflow, dag_sink_work, task_count, task_chunk
):
    """Add the special final PanDA task to the client workflow.

    Parameters
    ----------
    config : `lsst.ctrl.bps.BpsConfig`
        BPS configuration.
    generic_workflow : `lsst.ctrl.bps.GenericWorkflow`
        Generic workflow in which to find the final job.
    idds_client_workflow : `idds.workflowv2.workflow.Workflow`
        iDDS client representation of the workflow to which the final task
        is added.
    dag_sink_work : `list` [`idds.doma.workflowv2.domapandawork.DomaPanDAWork`]
        The work nodes in the client workflow which have no successors.
    task_count : `int`
        Count of PanDA tasks used when making unique names.
    task_chunk : `int`
        Count of chunk of a PanDA tasks used when making unique names.

    Returns
    -------
    files : `dict` [`str`, `str`]
        Files which need to be copied to a workflow staging area.

    Raises
    ------
    NotImplementedError
        Raised if final job in GenericWorkflow is itself a workflow.
    TypeError
        Raised if final job in GenericWorkflow is invalid type.
    """
    files = {}

    # If final job exists in generic workflow, create DAG final job
    final = generic_workflow.get_final()
    if final:
        if isinstance(final, GenericWorkflow):
            raise NotImplementedError("PanDA plugin does not support a workflow as the final job")

        if not isinstance(final, GenericWorkflowJob):
            raise TypeError(f"Invalid type for GenericWorkflow.get_final() results ({type(final)})")

        dag_final_work, files = _make_doma_work(
            config,
            generic_workflow,
            final,
            task_count,
            task_chunk,
        )
        pseudo_filename = "pure_pseudoinput+qgraphNodeId:+qgraphId:"
        dag_final_work.dependency_map.append(
            {"name": pseudo_filename, "submitted": False, "dependencies": []}
        )
        idds_client_workflow.add_work(dag_final_work)
        conditions = []
        for work in dag_sink_work:
            conditions.append(work.is_terminated)
        and_cond = AndCondition(conditions=conditions, true_works=[dag_final_work])
        idds_client_workflow.add_condition(and_cond)
    else:
        _LOG.debug("No final job in GenericWorkflow")
    return files


def convert_exec_string_to_hex(cmdline):
    """Convert the command line into hex representation.

    This step is currently involved because large blocks of command lines
    including special symbols passed to the pilot/container. To make sure
    the 1 to 1 matching and pass by the special symbol stripping
    performed by the Pilot we applied the hexing.

    Parameters
    ----------
    cmdline : `str`
        UTF-8 command line string

    Returns
    -------
    hex : `str`
        Hex representation of string
    """
    return binascii.hexlify(cmdline.encode()).decode("utf-8")


def add_decoder_prefix(config, cmd_line, distribution_path, files):
    """Compose the command line sent to the pilot from the functional part
    (the actual SW running) and the middleware part (containers invocation)

    Parameters
    ----------
    config : `lsst.ctrl.bps.BpsConfig`
        Configuration information
    cmd_line : `str`
        UTF-8 based functional part of the command line
    distribution_path : `str`
        URI of path where all files are located for distribution
    files : `tuple` [`dict` [`str`, `str`], `list` [`str`]]
        File names needed for a task (copied local, direct access)

    Returns
    -------
    decoder_prefix : `str`
        Full command line to be executed on the edge node
    """
    # Manipulate file paths for placement on cmdline
    files_plc_hldr = {}
    for key, pfn in files[0].items():
        if pfn.endswith("/"):
            files_plc_hldr[key] = os.path.basename(pfn[:-1])
            isdir = True
        else:
            files_plc_hldr[key] = os.path.basename(pfn)
            _, extension = os.path.splitext(pfn)
            isdir = os.path.isdir(pfn) or (key == "butlerConfig" and extension != "yaml")
        if isdir:
            # this is needed to make isdir function working
            # properly in ButlerURL instance on the egde node
            files_plc_hldr[key] += "/"
        _LOG.debug("files_plc_hldr[%s] = %s", key, files_plc_hldr[key])

    cmdline_hex = convert_exec_string_to_hex(cmd_line)
    _, runner_command = config.search("runnerCommand", opt={"replaceEnvVars": False, "expandEnvVars": False})
    runner_command = runner_command.replace("\n", " ")
    decoder_prefix = runner_command.replace(
        "_cmd_line_",
        str(cmdline_hex)
        + " ${IN/L} "
        + distribution_path
        + "  "
        + "+".join(f"{k}:{v}" for k, v in files_plc_hldr.items())
        + " "
        + "+".join(files[1]),
    )
    return decoder_prefix


def add_idds_work(config, generic_workflow, idds_workflow):
    """Convert GenericWorkflowJobs to iDDS work and add them to the iDDS
        workflow.

    Parameters
    ----------
    config : `lsst.ctrl.bps.BpsConfig`
        BPS configuration
    generic_workflow : `lsst.ctrl.bps.GenericWorkflow`
        Generic workflow containing jobs to convert.
    idds_workflow : `idds.workflowv2.workflow.Workflow`
        iDDS workflow to which the converted jobs should be added.

    Returns
    -------
    files_to_pre_stage : `dict` [`str`, `str`]
        Files that need to be copied to the staging area before submission.
    dag_sink_work : `list` [`idds.doma.workflowv2.domapandawork.DomaPanDAWork`]
        The work nodes in the client workflow which have no successors.
    task_count : `int`
        Number of tasks in iDDS workflow used for unique task names

    Raises
    ------
    RuntimeError
        If cannot recover from dependency issues after pass through workflow.
    """
    # Limit number of jobs in single PanDA task
    _, max_jobs_per_task = config.search("maxJobsPerTask", opt={"default": PANDA_DEFAULT_MAX_JOBS_PER_TASK})

    files_to_pre_stage = {}
    dag_sink_work = []  # Workflow sink nodes that need to be connected to final task
    job_to_task = {}
    job_to_pseudo_filename = {}
    task_count = 0  # Task number/ID in idds workflow used for unique name

    # To avoid dying due to optimizing number of times through workflow,
    # catch dependency issues to loop through again later.
    jobs_with_dependency_issues = {}

    # Assume jobs with same label share config values
    for job_label in generic_workflow.labels:
        _LOG.debug("job_label = %s", job_label)
        # Add each job with a particular label to a corresponding PanDA task
        # A PanDA task has a limit on number of jobs, so break into multiple
        # PanDA tasks if needed.
        job_count = 0  # Number of jobs in idds task used for task chunking
        task_chunk = 1  # Task chunk number within job label used for unique name
        work = None

        # Instead of changing code to make chunks up front and round-robin
        # assign jobs to chunks, for now keeping chunk creation in loop
        # but using knowledge of how many chunks there will be to set better
        # maximum number of jobs in a chunk for more even distribution.
        jobs_by_label = generic_workflow.get_jobs_by_label(job_label)
        num_chunks = -(-len(jobs_by_label) // max_jobs_per_task)  # ceil
        max_jobs_per_task_this_label = -(-len(jobs_by_label) // num_chunks)
        _LOG.debug(
            "For job_label = %s, num jobs = %s, num_chunks = %s, max_jobs = %s",
            job_label,
            len(jobs_by_label),
            num_chunks,
            max_jobs_per_task_this_label,
        )
        for gwjob in jobs_by_label:
            job_count += 1
            if job_count > max_jobs_per_task_this_label:
                job_count = 1
                task_chunk += 1

            if job_count == 1:
                # Create new PanDA task object
                task_count += 1
                work, files = _make_doma_work(config, generic_workflow, gwjob, task_count, task_chunk)
                files_to_pre_stage.update(files)
                idds_workflow.add_work(work)
                if generic_workflow.out_degree(gwjob.name) == 0:
                    dag_sink_work.append(work)

            pseudo_filename = _make_pseudo_filename(config, gwjob)
            job_to_pseudo_filename[gwjob.name] = pseudo_filename
            job_to_task[gwjob.name] = work.get_work_name()
            deps = []
            missing_deps = False
            for parent_job_name in generic_workflow.predecessors(gwjob.name):
                if parent_job_name not in job_to_task:
                    _LOG.debug("job_to_task.keys() = %s", job_to_task.keys())
                    missing_deps = True
                    break
                else:
                    deps.append(
                        {
                            "task": job_to_task[parent_job_name],
                            "inputname": job_to_pseudo_filename[parent_job_name],
                            "available": False,
                        }
                    )
            if not missing_deps:
                work.dependency_map.append({"name": pseudo_filename, "dependencies": deps})
            else:
                jobs_with_dependency_issues[gwjob.name] = work

    # If there were any issues figuring out dependencies through earlier loop
    if jobs_with_dependency_issues:
        _LOG.warning("Could not prepare workflow in single pass.  Please notify developers.")
        _LOG.info("Trying to recover...")
        for job_name, work in jobs_with_dependency_issues.items():
            deps = []
            for parent_job_name in generic_workflow.predecessors(job_name):
                if parent_job_name not in job_to_task:
                    _LOG.debug("job_to_task.keys() = %s", job_to_task.keys())
                    raise RuntimeError(
                        "Could not recover from dependency issues ({job_name} missing {parent_job_name})."
                    )
                deps.append(
                    {
                        "task": job_to_task[parent_job_name],
                        "inputname": job_to_pseudo_filename[parent_job_name],
                        "available": False,
                    }
                )
            pseudo_filename = job_to_pseudo_filename[job_name]
            work.dependency_map.append({"name": pseudo_filename, "dependencies": deps})
        _LOG.info("Successfully recovered.")

    return files_to_pre_stage, dag_sink_work, task_count
